package mccity.heroes.skills.rockslide;

import com.herocraftonline.heroes.Heroes;
import com.herocraftonline.heroes.api.SkillResult;
import com.herocraftonline.heroes.characters.Hero;
import com.herocraftonline.heroes.characters.skill.SkillConfigManager;
import com.herocraftonline.heroes.characters.skill.SkillType;
import com.herocraftonline.heroes.characters.skill.TargettedSkill;
import com.herocraftonline.heroes.util.Messaging;
import com.herocraftonline.heroes.util.Setting;
import com.herocraftonline.heroes.util.Util;
import org.bukkit.Bukkit;
import org.bukkit.Location;
import org.bukkit.Material;
import org.bukkit.World;
import org.bukkit.block.Block;
import org.bukkit.block.BlockFace;
import org.bukkit.configuration.ConfigurationSection;
import org.bukkit.craftbukkit.CraftWorld;
import org.bukkit.craftbukkit.entity.CraftPlayer;
import org.bukkit.craftbukkit.event.CraftEventFactory;
import org.bukkit.entity.LivingEntity;
import org.bukkit.entity.Player;
import org.bukkit.event.EventHandler;
import org.bukkit.event.EventPriority;
import org.bukkit.event.Listener;
import org.bukkit.event.block.BlockPlaceEvent;
import org.bukkit.event.entity.ItemSpawnEvent;
import org.bukkit.inventory.ItemStack;

import java.util.*;

public class SkillRockslide extends TargettedSkill implements Listener {

    private static final EnumSet<Material> FALL_FREE_MATS = EnumSet.of(Material.AIR, Material.WATER, Material.STATIONARY_WATER,
            Material.LAVA, Material.STATIONARY_LAVA, Material.FIRE);

    private final Set<WorldColumn> rockslideColumns = new HashSet<WorldColumn>();

    private String useFailedText;

    public SkillRockslide(Heroes plugin) {
        super(plugin, "Rockslide");
        setDescription("Drops down a Rockslide on the target.");
        setUsage("/skill rockslide <target>");
        setArgumentRange(0, 1);
        setIdentifiers("skill rockslide");
        setTypes(SkillType.EARTH, SkillType.HARMFUL, SkillType.SILENCABLE);

        Bukkit.getPluginManager().registerEvents(this, this.plugin);
    }

    @Override
    public void init() {
        super.init();
        useFailedText = SkillConfigManager.getRaw(this, "use-failed-text", "Cant use here");
    }

    public ConfigurationSection getDefaultConfig() {
        ConfigurationSection defaultConfig = super.getDefaultConfig();
        defaultConfig.set(Setting.COOLDOWN.node(), 30000);
        defaultConfig.set(Setting.MANA.node(), 30);
        defaultConfig.set(Setting.MAX_DISTANCE.node(), 10);
        defaultConfig.set(Setting.MAX_DISTANCE_INCREASE.node(), 0.1);
        defaultConfig.set(Setting.DURATION.node(), 2000);
        defaultConfig.set(Setting.DURATION_INCREASE.node(), 60);
        defaultConfig.set("radius", 2);
        defaultConfig.set("height", 3);
        defaultConfig.set("use-failed-text", "Cant use here");
        return defaultConfig;
    }

    public String getDescription(Hero hero) {
        StringBuilder descr = new StringBuilder(getDescription());

        double cdSec = SkillConfigManager.getUseSetting(hero, this, Setting.COOLDOWN, 30000, false) / 1000.0;
        if (cdSec > 0) {
            descr.append(" CD:");
            descr.append(Util.formatDouble(cdSec));
            descr.append("s");
        }

        int mana = SkillConfigManager.getUseSetting(hero, this, Setting.MANA, 30, false);
        if (mana > 0) {
            descr.append(" M:");
            descr.append(mana);
        }

        double distance = SkillConfigManager.getUseSetting(hero, this, Setting.MAX_DISTANCE.node(), 10, false) +
                SkillConfigManager.getUseSetting(hero, this, Setting.MAX_DISTANCE_INCREASE.node(), 0.1, false) * hero.getSkillLevel(this);
        if (distance > 0) {
            descr.append(" Dist:");
            descr.append(Util.formatDouble(distance));
        }

        return descr.toString();
    }

    @Override
    public SkillResult use(Hero hero, LivingEntity target, String[] args) {
        Location rockslideLoc = target.getLocation().getBlock().getLocation();
        rockslideLoc.add(0.0, Math.ceil(target.getEyeHeight()) + 1, 0.0);

        int radius = SkillConfigManager.getUseSetting(hero, this, "radius", 2, false);
        radius = Math.max(0, Math.min(radius, 3));

        // check can build at the bounds
        Player player = hero.getPlayer();
        boolean canBuild = canBuild(player, rockslideLoc.clone().add(-radius, 0, -radius));
        canBuild = canBuild && canBuild(player, rockslideLoc.clone().add(radius, 0 , radius));
        canBuild = canBuild && canBuild(player, rockslideLoc.clone().add(-radius, 0 , radius));
        canBuild = canBuild && canBuild(player, rockslideLoc.clone().add(radius, 0 , -radius));
        if (!canBuild) {
            Messaging.send(player, useFailedText);
            return SkillResult.FAIL;
        }

        int height = SkillConfigManager.getUseSetting(hero, this, "height", 3, false);
        height = Math.max(1, Math.min(height, 5));
        int x = rockslideLoc.getBlockX();
        int y = rockslideLoc.getBlockY();
        int z = rockslideLoc.getBlockZ();
        World world = rockslideLoc.getWorld();

        RockslideTask rockslideTask = new RockslideTask();
        for (int ix = x - radius; ix <= x + radius; ix++) {
            for (int iz = z - radius; iz <= z + radius; iz++) {
                for (int iy = y; iy < y + height; iy++) {
                    Location curLoc = new Location(world, ix, iy, iz);
                    Block curBlock = world.getBlockAt(curLoc);
                    if (curBlock.getType() == Material.AIR) {
                        if (!rockslideTask.tryAddFalling(curBlock)) {
                            Messaging.send(player, useFailedText);
                            return SkillResult.FAIL;
                        }
                    }
                }
            }
        }

        if (rockslideTask.hasBlocks()) {
            rockslideTask.launch();
        } else {
            Messaging.send(player, useFailedText);
            return SkillResult.FAIL;
        }

        // schedule rollback
        int durationMs = SkillConfigManager.getUseSetting(hero, this, Setting.DURATION.node(), 2000, false) +
                SkillConfigManager.getUseSetting(hero, this, Setting.DURATION_INCREASE.node(), 60, false) * hero.getSkillLevel(this);
        Bukkit.getScheduler().scheduleSyncDelayedTask(plugin, rockslideTask, durationMs / 50);

        return SkillResult.NORMAL;
    }

    private boolean canBuild(Player player, Location testLoc) {
        Block testLocBlock = testLoc.getWorld().getBlockAt(testLoc);

        BlockPlaceEvent bpe = CraftEventFactory.callBlockPlaceEvent(((CraftWorld) testLoc.getWorld()).getHandle(),
                ((CraftPlayer) player).getHandle(), testLocBlock.getState(),
                testLoc.getBlockX(), testLoc.getBlockY(), testLoc.getBlockZ());
        return !bpe.isCancelled() && bpe.canBuild();
    }

    @EventHandler(priority = EventPriority.NORMAL, ignoreCancelled = true)
    public void onItemSpawn(ItemSpawnEvent event) {
        if (rockslideColumns.isEmpty()) return;

        ItemStack dropStack = event.getEntity().getItemStack();
        if (dropStack.getType() == Material.GRAVEL && dropStack.getAmount() == 1) {
            Location dropLoc = event.getLocation();
            WorldColumn dropColumn = new WorldColumn(dropLoc.getBlockX(), dropLoc.getBlockZ(), dropLoc.getWorld());
            if (rockslideColumns.contains(dropColumn)) {
                event.setCancelled(true);
            }
        }
    }

    private class RockslideTask implements Runnable {

        private final List<Block> initialBlocks = new ArrayList<Block>();
        private final Set<WorldColumn> columns = new HashSet<WorldColumn>();

        public boolean tryAddFalling(Block blockPos) {
            initialBlocks.add(blockPos);

            WorldColumn wv2 = new WorldColumn(blockPos.getX(), blockPos.getZ(), blockPos.getWorld());
            if (rockslideColumns.contains(wv2)) {
                return false; // Rockslides should not cross each other
            } else {
                columns.add(wv2);
                return true;
            }
        }

        public boolean hasBlocks() {
            return !initialBlocks.isEmpty();
        }

        public void launch() {
            rockslideColumns.addAll(columns);
            for (Block curInitBlock : initialBlocks) {
                curInitBlock.setType(Material.GRAVEL);
            }
        }

        @Override
        public void run() {
            // rollback
            for (Block initialBlock : initialBlocks) {
                Block curBlock = initialBlock;
                while (true) {
                    Material curType = curBlock.getType();
                    if (curType == Material.GRAVEL) {
                        curBlock.setType(Material.AIR);
                    } else if (FALL_FREE_MATS.contains(curType)) {
                        curBlock = curBlock.getRelative(BlockFace.DOWN);
                        continue;
                    }
                    break;
                }
            }

            rockslideColumns.removeAll(columns);
        }
    }
}
