package mccity.heroes.skills.scent;

import com.herocraftonline.heroes.Heroes;
import com.herocraftonline.heroes.api.SkillResult;
import com.herocraftonline.heroes.characters.Hero;
import com.herocraftonline.heroes.characters.skill.ActiveSkill;
import com.herocraftonline.heroes.characters.skill.SkillConfigManager;
import com.herocraftonline.heroes.characters.skill.SkillType;
import com.herocraftonline.heroes.util.Messaging;
import com.herocraftonline.heroes.util.Setting;
import com.herocraftonline.heroes.util.Util;
import org.bukkit.configuration.ConfigurationSection;
import org.bukkit.entity.Entity;
import org.bukkit.entity.Player;

import java.util.List;

public class SkillScent extends ActiveSkill {

    private String playerDetectedText;
    private String nothingText;

    public SkillScent(Heroes plugin) {
        super(plugin, "Scent");
        setDescription("Try to determine whether there is a player near you.");
        setUsage("/skill scent");
        setArgumentRange(0, 0);
        setIdentifiers("skill scent");
        setTypes(SkillType.ILLUSION);
    }

    @Override
    public void init() {
        super.init();
        playerDetectedText = SkillConfigManager.getRaw(this, "player-detected-text", "You hear someone's footsteps");
        nothingText = SkillConfigManager.getRaw(this, "nothing-text", "Seems, there is nobody nearby");
    }

    public ConfigurationSection getDefaultConfig() {
        ConfigurationSection defaultConfig = super.getDefaultConfig();
        defaultConfig.set(Setting.COOLDOWN.node(), 60000);
        defaultConfig.set(Setting.MANA.node(), 50);
        defaultConfig.set(Setting.MAX_DISTANCE.node(), 30);
        defaultConfig.set(Setting.MAX_DISTANCE_INCREASE.node(), 0.4);
        defaultConfig.set("player-detected-text", "You hear someone's footsteps");
        defaultConfig.set("nothing-text", "Seems, there is nobody nearby");
        return defaultConfig;
    }

    public String getDescription(Hero hero) {
        StringBuilder descr = new StringBuilder(getDescription());

        double cdSec = SkillConfigManager.getUseSetting(hero, this, Setting.COOLDOWN, 60000, false) / 1000.0;
        if (cdSec > 0) {
            descr.append(" CD:");
            descr.append(Util.formatDouble(cdSec));
            descr.append("s");
        }

        int mana = SkillConfigManager.getUseSetting(hero, this, Setting.MANA, 50, false);
        if (mana > 0) {
            descr.append(" M:");
            descr.append(mana);
        }

        double distance = getMaxDistanceFor(hero);
        if (distance > 0) {
            descr.append(" Dist:");
            descr.append(Util.formatDouble(distance));
        }

        return descr.toString();
    }

    public double getMaxDistanceFor(Hero hero) {
        return SkillConfigManager.getUseSetting(hero, this, Setting.MAX_DISTANCE.node(), 30, false) +
                SkillConfigManager.getUseSetting(hero, this, Setting.MAX_DISTANCE_INCREASE.node(), 0.4, false) * hero.getSkillLevel(this);
    }

    @Override
    public SkillResult use(Hero hero, String[] args) {
        Player player = hero.getPlayer();
        double maxDistance = getMaxDistanceFor(hero);
        maxDistance = Math.max(1, Math.min(maxDistance, 100));

        List<Entity> nearbyEntities = player.getNearbyEntities(maxDistance, maxDistance, maxDistance);

        boolean found = false;
        for (Entity nearbyEntity : nearbyEntities) {
            if (nearbyEntity instanceof Player) {
                Player nearbyPlayer = (Player) nearbyEntity;
                if (nearbyPlayer.equals(player)) continue;

                Messaging.send(player, playerDetectedText);
                found = true;
                break;
            }
        }

        if (!found) {
            Messaging.send(player, nothingText);
        }

        return SkillResult.NORMAL;
    }
}
